/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
/*
 * The functions binding Relief together.
 */

#include <string.h>
#include <ncurses.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <assert.h>

#include "main.h"

#include "options.h"
#include "lib.h"
#include "curs_main.h"
#include "browser.h"
#include "keymap.h"
#include "misc.h"
#include "curs_menu.h"
#include "curs_lib.h"
#include "playlist.h"
#include "play.h"

#define BREAK_IF_NO_ENTRIES \
if (!menu->entries) { msg("There are no entries in the menu."); break; }

/* Return values of the switch functions */
#define NOT_FOUND (-1)
#define FOUND (0)

/* The screen row of the first menu entry on each page */
#define INDEX_OFFSET (1)
/* The maximum number of menu entries on each page */
#define PAGELEN (LINES - 3)

static void redraw_screen(void);
static int switch_generic(int cmd);
static int switch_main(int cmd);
static int switch_options(int cmd);
static int switch_browser_main(int cmd);
static int switch_browser_main_options(int cmd);
static int switch_browser(int cmd);
static int switch_browser_help_options(int cmd);

static bool option_changed = FALSE;
static bool playlist_changed = FALSE;
static bool continuous_flag = FALSE;
static bool random_flag = FALSE;
static char *playlist;		/* the path of the open playlist */
/* the current menu, the previous menu and the previous menu's previous menu */
static struct menu *menu;
static struct menu *prevmenu = NULL;
static struct menu *prevprevmenu = NULL;

/*
 * Initialize `menu' for the first time.
 *
 * if the `OPT_STARTUP_LOAD_DIR' option is true
 * 	load all songs in the directory specified
 *	no playlist will be open
 * else if the `OPT_STARTUP_PLAYLIST' option is true
 *	load the playlist specified
 * else
 *	there'll be no entries in the menu
 *	no playlist will be open
 *	
 */
void init_first_menu(void)
{
    menu = safe_malloc(sizeof(struct menu));

    menu->menu = MENU_MAIN;
    TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
    menu->make_entry = main_make_menu_entry;
    menu->search = playlist_search;
    menu->prev_main_search_op = NOP;
    menu->oldcurrent = menu->current = menu->top = 0;
    menu->offset = INDEX_OFFSET;
    menu->pagelen = PAGELEN;

    /* The `OPT_STARTUP_LOAD_DIR' option overrides `OPT_STARTUP_PLAYLIST' */
    if (is_path_option_set(OPT_STARTUP_LOAD_DIR)) {
	int n = add_dir(path_option_value(OPT_STARTUP_LOAD_DIR));
	if (n > 0) /* n = number of songs added */
	    menu->entries = n;
	playlist = NULL; /* Unnamed playlist */
    } else if (is_path_option_set(OPT_STARTUP_PLAYLIST)) {
        bool err;
	playlist = safe_strdup(path_option_value(OPT_STARTUP_PLAYLIST));
	first_entry = load_playlist(playlist, &(menu->entries), &err);
	if (first_entry == NULL && err == TRUE)
	    safe_free(playlist);
	/* else `playlist' is loaded, but there are no songs in it */
    } else {
	playlist = NULL; /* Unnamed playlist */
	menu->entries = 0;
    }
}

/*
 * Loop forever until the user triggers `CMD_QUIT'.
 * 1. Redraw the screen
 * 2. Get a command from the user
 * 3. Do the operation
 */
void menu_loop(void)
{
    int cmd;         /* command given by the user */
    int num_cmd_arg; /* numeric argument to the command */

    while (1) { /* forever */

	/* Hide the cursor (does not work on all terminals) */
	(void) curs_set(0);

	redraw_screen();

	/* Get a command from the user */
	menu_get_command(menu, &cmd, &num_cmd_arg);

	/* Clear the message line, in case there's a message on it (relating
	   to the previous command) */
	CLEARLINE(MSG_LINE);

        /* Unhide the cursor for the operation to be executed */
	(void) curs_set(1);

	refresh();

	/* These functions do the operation and return `FOUND' if `cmd'
	   is valid for them. */

	if (switch_generic(cmd) == FOUND)
	    continue;

	if (menu->menu == MENU_MAIN)
	    if (switch_main(cmd) == FOUND)
		continue;

	if (menu->menu == MENU_OPTIONS)
	    if (switch_options(cmd) == FOUND)
		continue;

	if (menu->menu == MENU_MAIN || menu->menu == MENU_BROWSER)
	    if (switch_browser_main(cmd) == FOUND)
		continue;

	if (menu->menu == MENU_MAIN || menu->menu == MENU_BROWSER
	    || menu->menu == MENU_OPTIONS)
	    if (switch_browser_main_options(cmd) == FOUND)
		continue;

	if (menu->menu == MENU_BROWSER)
	    if (switch_browser(cmd) == FOUND)
		continue;

	if (menu->menu == MENU_BROWSER || menu->menu == MENU_HELP ||
	    menu->menu == MENU_OPTIONS)
	    if (switch_browser_help_options(cmd) == FOUND)
		continue;

	/* menu_get_command() returns only valid commands,
	   so if we reach this far, the command exists in `keymap',
	   but there are no case for it in the switch functions */
	assert(0);
    }
}

/*
 * Redraw the screen.
 */
static void redraw_screen(void)
{
    if (IS_FLAG_ON(menu->redraw, REDRAW_FULL)) {
	(void) clear();
	CLEAR_ALL_FLAGS(menu->redraw);
	TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	TURN_ON_FLAG(menu->redraw, REDRAW_CMD_REF);
    }
    if (IS_FLAG_ON(menu->redraw, REDRAW_STATUS)) {
	menu_redraw_status(menu, playlist, continuous_flag, random_flag);
	TURN_OFF_FLAG(menu->redraw, REDRAW_STATUS);
    }
    if (IS_FLAG_ON(menu->redraw, REDRAW_CMD_REF)) {
	menu_redraw_cmd_ref(menu);
	TURN_OFF_FLAG(menu->redraw, REDRAW_CMD_REF);
    }

    /* Update `menu->top' (since some functions don't bother doing it) */
    if (menu->entries) {
	if ((menu->current >= menu->top + menu->pagelen) ||
	    (menu->current < menu->top)) {
	    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	}
	/* Maximum one of these two loops will be run. */
	while (menu->current >= menu->top + menu->pagelen)
	    menu->top++;
	/* menu->top will never be < 0 since menu->current is >= 0 */
	while (menu->current < menu->top)
	    menu->top--;
    }

    /* Only one of those two flags should be on at the same time. */    
    assert(!(IS_FLAG_ON(menu->redraw, REDRAW_MOTION) &&
	     IS_FLAG_ON(menu->redraw, REDRAW_CURRENT)));

    /* If both `REDRAW_INDEX' and either `REDRAW_MOTION' or `REDRAW_CURRENT'
       are on, menu_redraw_index() is sufficient. */
    if (IS_FLAG_ON(menu->redraw, REDRAW_INDEX))
	menu_redraw_index(menu);
    else if (IS_FLAG_ON(menu->redraw, REDRAW_MOTION))
	menu_redraw_motion(menu);
    else if (IS_FLAG_ON(menu->redraw, REDRAW_CURRENT))
	menu_redraw_current(menu);

    CLEAR_ALL_FLAGS(menu->redraw);
    /* `oldcurrent' is used by the redrawing functions above. By doing this
       assignment here, the 10+ user commands triggering functions which change
       `current' do not need to worry about `oldcurrent'. */
    menu->oldcurrent = menu->current;

    /* In the options menu, we display a description of the current
       option on the msg line. We always start the description with
       a single space, so we can check to see if the text on the msg
       line is an option description. If it is, we can overwrite it. */
    if (menu->menu == MENU_OPTIONS) {
	/* Extract only the character, not the attributes. */
	if ((mvinch(MSG_LINE, 0) & A_CHARTEXT) == ' ') {
	    if (is_a_non_existing_option(menu->current))
		msg("%*s", COLS, " ");
	    else
		msg(" %s", option_description_string(menu->current));
	}
    }
    /* Move the cursor as much out of sight as possible (for terminals
       not supporting curs_set()) */
    if (menu->entries)
	move(menu->current - menu->top + menu->offset, COLS - 1);
    else
	move(1, COLS - 1);

    refresh();

    assert(menu->current >= 0);
    assert(menu->entries >= 0);
    assert(menu->entries ? menu->current <= menu->entries - 1 : TRUE);
}

/*
 * If `cmd' is a generic command (valid in all menus), do the command
 * and return `FOUND', else return `NOT_FOUND'.
 */
static int switch_generic(int cmd)
{
    switch (cmd) {

    case CMD_GO_TO:
	cmd_go_to(menu);
	break;
    case CMD_TOP_OF_PAGE:
	cmd_top_of_page(menu);
	break;
    case CMD_MIDDLE_OF_PAGE:
	cmd_middle_of_page(menu);
	break;
    case CMD_BOTTOM_OF_PAGE:
	cmd_bottom_of_page(menu);
	break;
    case CMD_FIRST_ENTRY:
	cmd_first_entry(menu);
	break;
    case CMD_LAST_ENTRY:
	cmd_last_entry(menu);
	break;
    case CMD_PAGE_DOWN:
	cmd_page_down(menu);
	break;
    case CMD_PAGE_UP:
	cmd_page_up(menu);
	break;
    case CMD_SCROLL_UP:
	cmd_scroll_up(menu);
	break;
    case CMD_SCROLL_DOWN:
	cmd_scroll_down(menu);
	break;
    case CMD_NEXT_ENTRY:
	cmd_next_entry(menu);
	break;
    case CMD_PREVIOUS_ENTRY:
	cmd_previous_entry(menu);
	break;
    case CMD_HALF_PAGE_DOWN:
	cmd_half_page_down(menu);
	break;
    case CMD_HALF_PAGE_UP:
	cmd_half_page_up(menu);
	break;
    case CMD_REDRAW_SCREEN:
	cmd_redraw_screen(&(menu->redraw));
	break;
    default:
	return NOT_FOUND;
    }

    return FOUND;
}

/*
 * If `cmd' is a valid command for `MENU_MAIN', and not a generic
 * command, do the command and return `FOUND', else return `NOT_FOUND'.
 */
static int switch_main(int cmd)
{
    switch (cmd) {

    case CMD_MOVE_SONG_TO:
	cmd_move_song_to(menu, &playlist_changed);
	break;
    case CMD_MOVE_SONG_TO_TOP_OF_PAGE:
	cmd_move_song_to_top_of_page(menu, &playlist_changed);
	break;
    case CMD_MOVE_SONG_TO_MIDDLE_OF_PAGE:
	cmd_move_song_to_middle_of_page(menu, &playlist_changed);
	break;
    case CMD_MOVE_SONG_TO_BOTTOM_OF_PAGE:
	cmd_move_song_to_bottom_of_page(menu, &playlist_changed);
	break;
    case CMD_MOVE_SONG_FIRST:
	cmd_move_song_first(menu, &playlist_changed);
	break;
    case CMD_MOVE_SONG_LAST:
	cmd_move_song_last(menu, &playlist_changed);
	break;
    case CMD_MOVE_SONG_DOWN:
	cmd_move_song_down(menu, &playlist_changed);
	break;
    case CMD_MOVE_SONG_UP:
	cmd_move_song_up(menu, &playlist_changed);
	break;
    case CMD_ADD_DIRECTORY:
	interactive_add_dir(&menu, &prevmenu, &playlist_changed);
	break;
    case CMD_ADD_SONG:
	interactive_add_song(&menu, &prevmenu, &playlist_changed);
	break;
    case CMD_APPEND_PLAYLIST:
	interactive_append_playlist(menu, &playlist_changed);
	break;
    case CMD_CLEAR_PLAYLIST:
	free_playlist(&first_entry);
	playlist_changed = TRUE;
	menu->oldcurrent = menu->current = 0;
	menu->entries = menu->top = 0;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	break;

    case CMD_OPEN_PLAYLIST:
	interactive_load_playlist(menu, &playlist,
				  &playlist_changed);
	break;

    case CMD_TOGGLE_ALL_SONGS_OFF:
	BREAK_IF_NO_ENTRIES;
	toggle_all_entries(0);
	/* Assume the playlist was changed */
	playlist_changed = TRUE;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	break;

    case CMD_OPTIONS:
	prevmenu = menu;
	menu = safe_malloc(sizeof(struct menu));
	menu->menu = MENU_OPTIONS;
	menu->search = NULL;
	menu->prev_main_search_op = NOP;
	menu->offset = prevmenu->offset;
	menu->pagelen = prevmenu->pagelen;
	menu->entries = number_of_options();
	menu->oldcurrent = menu->current = menu->top = 0;
	menu->make_entry = options_make_menu_entry;
	TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
	break;

    case CMD_PLAY_SONG:
	BREAK_IF_NO_ENTRIES;
	play_song(menu->current);
	TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
	break;

    case CMD_PLAY_TOGGLED_SONGS:
	BREAK_IF_NO_ENTRIES;
	play_toggled_songs(continuous_flag, random_flag);
	TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
	break;

    case CMD_QUIT:
	quit(menu, playlist, playlist_changed);
	/* If we reach this far, `OPT_CONFIRM_QUIT_CMD' is true
	   and the user chose not to quit */
	break;

    case CMD_REMOVE_SONG:
	remove_entry(menu, &playlist_changed);
	break;

    case CMD_SAVE_PLAYLIST:
	interactive_save_playlist(menu, &playlist, &playlist_changed);
	break;

    case CMD_SAVE_PLAYLIST_AS:
	interactive_save_playlist_as(menu, &playlist, &playlist_changed);
	break;

    case CMD_TOGGLE_CONTINUOUS_FLAG:
	continuous_flag = ~continuous_flag;
	TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
	break;

    case CMD_TOGGLE_SONG:
	BREAK_IF_NO_ENTRIES;
	toggle_entry(menu->current);
	playlist_changed = TRUE;
	TURN_ON_FLAG(menu->redraw, REDRAW_CURRENT);
	break;

    case CMD_TOGGLE_RANDOM_FLAG:
	random_flag = ~random_flag;
	TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
	break;

    case CMD_WRITE_SONG_NAMES:
	interactive_write_song_names_to_file(menu);
	break;

    case CMD_TOGGLE_ALL_SONGS_ON:
	BREAK_IF_NO_ENTRIES;
	toggle_all_entries(1);
	/* Assume the playlist was changed */
	playlist_changed = TRUE;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	break;

    default:
	return NOT_FOUND;
    }

    return FOUND;
}

/*
 * If `cmd' is a valid command for `MENU_OPTIONS', and not a generic
 * command, do the command and return `FOUND', else return `NOT_FOUND'.
 */
static int switch_options(int cmd)
{
    switch (cmd) {

    case CMD_EDIT_OPTION:
	edit_option(menu->current, &(menu->redraw), &option_changed);
	break;

    case CMD_RESET_ALL_OPTIONS:
	reset_all_options();
	msg("All options reset to their default values.");
	option_changed = TRUE;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	break;
	      
    case CMD_TOGGLE_OPTION:
	/* see explanation at the head of the function definition */
	if (is_a_non_existing_option(menu->current))
	    break;
	if (!is_a_boolean_option(menu->current)) {
	    msg("Only boolean variables can be toggled.");
	    break;
	}
	toggle_boolean_option(menu->current);
	option_changed = TRUE;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	break;

    default:
	return NOT_FOUND;
    }

    return FOUND;
}

/*
 * If `cmd' is a valid command for `MENU_BROWSER' and `MENU_MAIN',
 * and not a generic command, do the command and
 * return `FOUND', else return `NOT_FOUND'.
 */
static int switch_browser_main(int cmd)
{
    switch (cmd) {

    case CMD_CHANGE_DIRECTORY:
	if (menu->menu == MENU_BROWSER)
	    browser_chdir(menu);
	else if (menu->menu == MENU_MAIN)
	    main_chdir(menu);
	else
	    assert(0);
	break;

    case CMD_SORT_REVERSE:
	BREAK_IF_NO_ENTRIES;
	if (menu->menu == MENU_MAIN) {
	    reverse_sort_playlist();
	    /* Assume the playlist was changed */
	    playlist_changed = TRUE;
	} else if (menu->menu == MENU_BROWSER) {
	    browser_reverse_sort_dir_entries();
	} else
	    assert(0);
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	break;

    case CMD_SORT:
	BREAK_IF_NO_ENTRIES;
	if (menu->menu == MENU_MAIN) {
	    sort_playlist();
	    /* Assume the playlist was changed */
	    playlist_changed = TRUE;
	} else if (menu->menu == MENU_BROWSER) {
	    browser_sort_dir_entries();
	} else
	    assert(0);
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	break;

    case CMD_REGEXP_SEARCH_BACKWARD:
    case CMD_REGEXP_SEARCH_FORWARD:
    case CMD_REPEAT_LAST_SEARCH:
    case CMD_REPEAT_LAST_SEARCH_OPPOSITE:
	assert(menu->search);
	menu->search(menu, cmd);
	break;

    default:
	return NOT_FOUND;
    }

    return FOUND;
}

/*
 * If `cmd' is a valid command for `MENU_BROWSER', `MENU_MAIN' and
 * `MENU_OPTIONS', and not a generic command, do the command and
 * return `FOUND', else return `NOT_FOUND'.
 */
static int switch_browser_main_options(int cmd)
{
    switch (cmd) {

    case CMD_HELP:
	if (menu->menu == MENU_MAIN) {
	    prevmenu = menu;
	    menu = safe_malloc(sizeof(struct menu));
	    menu->offset = prevmenu->offset;
	    menu->pagelen = prevmenu->pagelen;
	    menu->make_entry = menu_main_help_make_entry;
	} else {
	    prevprevmenu = safe_malloc(sizeof(struct menu));
	    *prevprevmenu = *prevmenu;
	    prevmenu = menu;
	    menu = safe_malloc(sizeof(struct menu));
	    menu->offset = prevmenu->offset;
	    menu->pagelen = prevmenu->pagelen;
	    if (prevmenu->menu == MENU_BROWSER) {
		menu->make_entry = menu_browser_help_make_entry;
	    } else if (prevmenu->menu == MENU_OPTIONS) {
		menu->make_entry = menu_options_help_make_entry;
	    } else
		assert(0);
	}
	menu->entries = number_of_menu_help_entries(prevmenu->menu);
	menu->search = NULL;
	menu->prev_main_search_op = NOP;
	menu->menu = MENU_HELP;
	menu->oldcurrent = menu->current = menu->top = 0;
	TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
	break;

    default:
	return NOT_FOUND;
    }

    return FOUND;
}

/*
 * If `cmd' is a valid command for `MENU_BROWSER', and not a generic
 * command, do the command and return `FOUND', else return `NOT_FOUND'.
 */
static int switch_browser(int cmd)
{
    switch (cmd) {

    case CMD_BROWSE_DIRECTORY:
	browser_open_directory(menu);
	break;

	/* nb! the use of `prevmenu' here is kind of ugly */
    case CMD_ADD_CURRENT_DIRECTORY:
    case CMD_ADD_CURRENT_SONG:
	if (cmd == CMD_ADD_CURRENT_DIRECTORY) {
	    int i = browser_add_dir(menu->current);
	    if (i > 0) {
		playlist_changed = TRUE;
		prevmenu->entries += i;
	    } else
		break;
	} else {
	    if (browser_add_song(menu->current) == 0) {
		playlist_changed = TRUE;
		prevmenu->entries++;
	    } else
		break;
	}

	browser_free_dir_list();
	safe_free(menu);
	menu = prevmenu;
	TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
	break;

    default:
	return NOT_FOUND;
    }

    return FOUND;
}

/*
 * If `cmd' is a valid command for `MENU_HELP' and `MENU_OPTIONS', and not
 * a generic command, do the command and return `FOUND', else return
 * `NOT_FOUND'.
 */
static int switch_browser_help_options(int cmd)
{
    switch (cmd) {

    case CMD_PREVIOUS_MENU:
	if (prevmenu->menu == MENU_MAIN) {
	    struct menu *tmpmenu = prevmenu;
	    prevmenu = menu;
	    menu = tmpmenu;
	    if (prevmenu->menu == MENU_OPTIONS && option_changed) {
		menu->offset = INDEX_OFFSET;
		menu->pagelen = PAGELEN;
		if (yes_or_no(YES, "Save changes?") == YES)
		    save_options(PersonalRc);
		option_changed = FALSE;
	    }
	    safe_free(prevmenu);
	} else if (menu->menu == MENU_HELP) {
	    free(menu);
	    menu = prevmenu;
	    prevmenu = prevprevmenu;
	    prevprevmenu = NULL;
	} else if (menu->menu == MENU_BROWSER) {
	    browser_free_dir_list();
	    safe_free(menu);
	    menu = prevmenu;
	    TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
	    break;
	} else
	    assert(0);

	TURN_ON_FLAG(menu->redraw, REDRAW_FULL);
	break;

    default:
	return NOT_FOUND;
    }

    return FOUND;
}
